//+------------------------------------------------------------------+
//|                                          Tester-512_Test_001.mq5 |
//|                                       Copyright 2012, MetaDriver |
//|                                            MetaDriver@rambler.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2012, MetaDriver"
#property link      "MetaDriver@rambler.ru"
#property version   "1.00"

#property script_show_inputs

input int OCL_Device = 0;

#include "Genom16x7x3.mqh"
#include "ArrayGenom.mqh"


#define CountInd   16
#define CountBars  50000
#define CountPass  512

g16x7x3     NETs[CountPass];
float       Inds[CountBars][CountInd];
float       Result[CountPass];
float       Price[CountBars];

CArrayGenom AG;

int cl_Ctx=INVALID_HANDLE;
int cl_Prg=INVALID_HANDLE;
int cl_Krn=INVALID_HANDLE;
int cl_Mem_NET=INVALID_HANDLE;
int cl_Mem_Inds=INVALID_HANDLE;
int cl_Mem_Price=INVALID_HANDLE;
int cl_Mem_Result=INVALID_HANDLE;
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
   if(CountInd!=8 && CountInd!=16)
     {
      Alert("Error : CountInd mast be == 8 or == 16 only !"); return;
     }
   WriteCL_Program();
   InitData();
   long GpuTime, CpuTime;
   if(!GpuInit()) return; else Print("OpenCL init OK!");
   
   long st=GetTickCount();
   gpuCalculate(); 
   Print("GPU time = ",GpuTime=GetTickCount()-st," ms");
   OutResult("Gpu");
   
   GpuDeinit();
   
   st=GetTickCount();
   cpuCalculate(); 
   Print("CPU time = ",CpuTime=GetTickCount()-st," ms");
   OutResult("Cpu");
   Print("CpuTime/GpuTime = ",CpuTime*1./GpuTime);
  }
//+------------------------------------------------------------------+
void InitData()
  {
   srand(GetTickCount());
   Price[1]=float(rand()*1.0/3276700.0+1.5);
   for(int i=1;i<CountBars;i++)
     {
      Price[i]=float(Price[i-1]+rand()*1.0/3276700.0);
     }
   for(int i=0;i<CountBars;i++)
     {
      for(int j=0;j<8;j++)
        {
         Inds[i][j]=float(rand()*1.0/32767.0-0.5);
        }
     }
   for(int i=0;i<CountPass;i++)
     {
      double len=0,s=0,temp[CountInd];
      for(int j=0;j<CountInd;j++)
        {
         s=temp[j]=float(rand()*1.0/32767.0-0.5);
         len+=s*s;
        }
      len=MathSqrt(len);
     }
    for(int i=0;i<CountPass;i++)
      {
       CGenom16x7x3 *item = new CGenom16x7x3(3);
       item.RandomInit();
       AG.Add(item);
      }
   for(int i=0; i<CountPass; i++)  { ((CGenom16x7x3*)AG.At(i)).CopyTo(NETs[i]);}
  }
//+------------------------------------------------------------------+
//float nets[sizeof(NETs)];
float inds[sizeof(Inds)];
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void gpuCalculate()
  {
   const uint Offs[1]  = {0};
   const uint Works[1] = {CountPass};
   CLSetKernelArgMem(cl_Krn,0,cl_Mem_NET);
   CLSetKernelArgMem(cl_Krn,1,cl_Mem_Inds);
   CLSetKernelArgMem(cl_Krn,2,cl_Mem_Price);
   CLSetKernelArgMem(cl_Krn,3,cl_Mem_Result);
   for(int i=0; i<CountPass; i++)  { ((CGenom16x7x3*)AG.At(i)).CopyTo(NETs[i]);}
   CLBufferWrite(cl_Mem_NET,NETs);
   for(int i=0; i<CountBars; i++) {for(int j=0; j<CountInd; j++) {inds[i*CountInd+j]=Inds[i][j];}}
   CLBufferWrite(cl_Mem_Inds,inds);
   CLBufferWrite(cl_Mem_Price,Price);
   CLExecute(cl_Krn,1,Offs,Works);
   CLBufferRead(cl_Mem_Result,Result);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void cpuCalculate()
  {
   for(int k=0;k<CountPass;k++)
     {
      float Res=0;
      for(int j=1;j<CountBars;j++)
        {
         Res+=cpuGetOneNETRes(j,NETs[k])*(Price[j]-Price[j-1]);
        }
      Result[k]=Res;
     }
  }
float sign(float x) {return (x<0)?-1.f:1.f;}
float cpuGetOneNETRes(int j, const g16x7x3 &NET)
{
  float  res7[7]={0,0,0,0,0,0,0};
  for(int i=0;i<7;i++)
    { for(int ii=0;ii<16;ii++) 
       { res7[i]+=Inds[j][ii]*NET.L_1[i][ii]; } }
  float  res3[3]={0,0,0};
  for(int i=0;i<3;i++)
    { for(int ii=0;ii<7;ii++) { res3[i]+=sign(res7[ii])*NET.L_23[i][ii]; } }
  float Res = 0;
  for(int i=0;i<3;i++) { Res+=sign(res3[i])*NET.L_23[i][7]; }
  return sign(Res);
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OutResult(string S)
  {
   Print("unt inticators = ", CountInd, "; Count history bars = ", CountBars,
         "; Count pass = ", CountPass);
   int Max=ArrayMaximum(Result);
   Print("Result on "+S+" Result==",Result[Max]," at ",Max," pass");
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool GpuInit()
  {
   ResetLastError();
//   if((cl_Ctx=CLContextCreate(CL_DEVICE_ANY))==INVALID_HANDLE)
//   if((cl_Ctx=CLContextCreate(0))==INVALID_HANDLE)
   if((cl_Ctx=CLContextCreate(OCL_Device))==INVALID_HANDLE) 
     {
      Print("OpenCL not found.");
      return false;
      //} else { Print("CLContextCreate() .. OK!");
     }
   if((cl_Prg=CLProgramCreate(cl_Ctx,Get_cl_Source()))==INVALID_HANDLE)
     {
      GpuDeinit();
      Print("OpenCL program create failed.");
      Print("GetLastError returned .. ",GetLastError());
      return false;
      //} else { Print("CLProgramCreate() .. OK!");
     }
   if((cl_Krn=CLKernelCreate(cl_Prg,"Tester_Test"))==INVALID_HANDLE)
     {
      GpuDeinit();
      Print("OpenCL kernel cl_Krn create failed.");
      Print("GetLastError returned .. ",GetLastError());
      switch(GetLastError())
        {
         case ERR_OPENCL_INVALID_HANDLE:
            Print("Err ==    program OpenCL");
            break;
         case ERR_INVALID_PARAMETER:
            Print("Err ==   ");
            break;
         case ERR_OPENCL_TOO_LONG_KERNEL_NAME:
            Print("Err ==     127 ");
            break;
         case ERR_OPENCL_KERNEL_CREATE:
            Print("Err ==       OpenCL.");
            break;
         default: Print("Err ==  .");
        }
      return false;
      //} else { Print("CLKernelCreate() .. OK!");
     }
   if((cl_Mem_NET=CLBufferCreate(cl_Ctx,CountPass*sizeof(g16x7x3),
      CL_MEM_WRITE_ONLY))==INVALID_HANDLE)
     {
      GpuDeinit();
      Print("OpenCL cl_Mem_NET buffer create failed.");
      return false;
      //} else { Print("CLBufferCreate(..cl_Mem_NET) .. OK!"); 
     }
   if((cl_Mem_Inds=CLBufferCreate(cl_Ctx,CountInd*CountBars*sizeof(float),
      CL_MEM_WRITE_ONLY))==INVALID_HANDLE)
     {
      GpuDeinit();
      Print("OpenCL cl_Mem_Inds buffer create failed.");
      return false;
      //} else { Print("CLBufferCreate(..cl_Mem_Inds) .. OK!"); 
     }
   if((cl_Mem_Price=CLBufferCreate(cl_Ctx,CountBars*sizeof(float),
      CL_MEM_WRITE_ONLY))==INVALID_HANDLE)
     {
      GpuDeinit();
      Print("OpenCL cl_Mem_Price buffer create failed.");
      return false;
      //} else { Print("CLBufferCreate(..cl_Mem_Price) .. OK!");
     }
   if((cl_Mem_Result=CLBufferCreate(cl_Ctx,CountPass*sizeof(float),
      CL_MEM_READ_ONLY))==INVALID_HANDLE)
     {
      GpuDeinit();
      Print("OpenCL cl_Mem_Result buffer create failed.");
      return false;
      //} else { Print("CLBufferCreate(..cl_Mem_Result) .. OK!");
     }
   return true;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool valid_h(int handle) {return handle!=INVALID_HANDLE;}
void GpuDeinit()
  {
   if (valid_h(cl_Mem_Result)) CLBufferFree(cl_Mem_Result);
   if (valid_h(cl_Mem_Price)) CLBufferFree(cl_Mem_Price);
   if (valid_h(cl_Mem_Inds)) CLBufferFree(cl_Mem_Inds);
   if (valid_h(cl_Mem_NET)) CLBufferFree(cl_Mem_NET);
   if (valid_h(cl_Krn)) CLKernelFree(cl_Krn);
   if (valid_h(cl_Prg)) CLProgramFree(cl_Prg);
   if (valid_h(cl_Ctx)) CLContextFree(cl_Ctx);
  }
//+------------------------------------------------------------------+
string i2s(int x){return IntegerToString(x);}
//+------------------------------------------------------------------+
string Get_cl_Source()
  {
   string S=
     "#define  CountInd   "+i2s(CountInd)+"            \r\n"
     "#define  CountBars  "+i2s(CountBars)+"            \r\n"
     "#define  CountPass  "+i2s(CountPass)+"             \r\n"
     "#define  dot8(a,b)  (dot(a.hi,b.hi)+dot(a.lo,b.lo)) \r\n"
     "#define  dot16(a,b) "
     "(dot(a.lo.lo,b.lo.lo)+dot(a.lo.hi,b.lo.hi)+"
     "dot(a.hi.lo,b.hi.lo)+dot(a.hi.hi,b.hi.hi))              \r\n"
     "                                                         \r\n"
     "typedef struct                                            \r\n"
     " {                                                        \r\n"
     "  float16 L1[7];                                          \r\n"
     "  float8  L23[3];                                         \r\n"
     "  float8  reserved;                                       \r\n"
     " } g16x7x3;                                               \r\n"
     "                                                          \r\n"
     " __kernel void Tester_Test(                               \r\n"
     "       global g16x7x3 *NET,                               \r\n"
     "       global float16 *Inds,                              \r\n"
     "       global float   *Price,                             \r\n"
     "       global float   *Result)                            \r\n"
     " {                                                        \r\n"
     "    int  pass = get_global_id(0);                         \r\n"
     "    g16x7x3 a = NET[pass];                                \r\n"
     "                                                          \r\n"
     "    float sum=0;                                          \r\n"
     "                                                          \r\n"
     "    for (int i=1; i<CountBars; i++)                       \r\n"
     "     {                                                    \r\n"
     "                                                          \r\n"
     "      float16 b = Inds[i];                                \r\n"
     "                                                          \r\n"
     "      float8 L1_sum = {                                   \r\n"
     "            dot16(a.L1[0],b),                             \r\n"
     "            dot16(a.L1[1],b),                             \r\n"
     "            dot16(a.L1[2],b),                             \r\n"
     "            dot16(a.L1[3],b),                             \r\n"
     "            dot16(a.L1[4],b),                             \r\n"
     "            dot16(a.L1[5],b),                             \r\n"
     "            dot16(a.L1[6],b),                             \r\n"
     "            0 };                                          \r\n"
     "      L1_sum = sign(L1_sum);                              \r\n"
     "      float3 L2_sum = {                                   \r\n"
     "            dot8(L1_sum,a.L23[0]),                        \r\n"
     "            dot8(L1_sum,a.L23[1]),                        \r\n"
     "            dot8(L1_sum,a.L23[2])                         \r\n"
     "            };                                            \r\n"
     "      L2_sum = sign(L2_sum);                              \r\n"
     "      float3 L3 = {                                       \r\n"
     "            a.L23[0].s7,                                  \r\n"
     "            a.L23[1].s7,                                  \r\n"
     "            a.L23[2].s7                                   \r\n"
     "            };                                            \r\n"
     "      sum += sign(dot(L2_sum,L3)) * (Price[i]-Price[i-1]);\r\n"
     "                                                          \r\n"
     "                                                         \r\n"
     "     } // for i                                         \r\n"
     "                                                       \r\n"
     "                                                     \r\n"
     "    Result[pass]= sum;                            \r\n"
     " }                                           \r\n"
     ;
   return S;
  }
//+------------------------------------------------------------------+
//+------------------------------------------------------------------+
void    WriteCL_Program()
  {
   int h=FileOpen("cl_ParallelTester_00-02-(16x7x3).cl",FILE_WRITE|FILE_TXT|FILE_ANSI);
   string S=Get_cl_Source();
// StringReplace(S,"\r\n",)
   FileWrite(h,S);
   FileClose(h);
  }
//+------------------------------------------------------------------+
